"""Module with the definition of the base class for both the thermo-physical
and the thermo-chemical properties, i.e., :class:`.PropertyInterface`."""
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Union
from numpy import nan
from scipy.optimize import minimize_scalar
from scipy.constants import atm
from .._decorators import range_warning


class PropertyInterface(ABC):
    """
    Abstract class that defines the interface for both the thermo-physical
    and the thermo-chemical properties.
    Derived classes must override the following attributes:

        - :func:`~PropertyInterface.correlation`
        - :attr:`~.PropertyInterface.range`
        - :attr:`~.PropertyInterface.units`
        - :attr:`~.PropertyInterface.long_name`
        - :attr:`~.PropertyInterface.description`

    Even the following attributes can be overridden, but not mandatorily:

        - :func:`~PropertyInterface.initialization_helper`: override this \
          method if the property correlation is particularly *difficult* to \
          solve in terms of temperature (see \
          :func:`lbh15.properties.lead_properties.p_s.initialization_helper`).\
          If not overridden, it is ignored.
        - :attr:`~.PropertyInterface.name`: override this member to give \
          the property a custom name, otherwise the class name is used.
        - :attr:`~.PropertyInterface.correlation_name`: override this member \
          to give the correlation a name different from the default one.
        - :attr:`~.PropertyInterface.is_injective`: override this member only \
          in case the correlation function is not injective by returning the
          `False` value, otherwise it is considered injective.

    Derived classes are not allowed having either the name starting with
    :attr:`__` or overriding the :attr:`~.PropertyInterface.name` attribute
    for providing a name starting with :attr:`__`.
    """
    def __init__(self):
        self.__min: float = -nan
        self.__max: float = nan
        self.__T_at_min: float = -nan
        self.__T_at_max: float = nan

    def compute_bounds(self) -> None:
        """
        Computes the bounds of the property within the validity range, i.e.,
        the minimum and the maximum values of the correlation function inside
        the validity range, together with the corresponding temperature values.
        If this method is not invoked, the default values are `-nan` for the
        minimum value of the property and for its corresponding temperature,
        and `nan` for the maximum value of the property and for its
        corresponding temperature.
        The bounding values are computed using the
        :func:`scipy.optimize.minimize_scalar` function by adopting the
        "Bounded" method (for more details, please refer to the *scipy*
        documentation).

        Returns
        -------
        None
        """
        res = minimize_scalar(self.correlation,
                              bounds=self.range,
                              method="Bounded",
                              options={'xatol': 1e-10})
        if res.success:
            self.__min = res.fun
            self.__T_at_min = res.x
        else:
            raise RuntimeError("Unable to find the minimum point: "
                               + res.message)

        def corr_opposite(T: float) -> float:
            return -self.correlation(T)

        res = minimize_scalar(corr_opposite,
                              bounds=self.range,
                              method="Bounded",
                              options={'xatol': 1e-10})
        if res.success:
            self.__max = -res.fun
            self.__T_at_max = res.x
        else:
            raise RuntimeError("Unable to find the maximum point: "
                               + res.message)

    def initialization_helper(self,
                              property_value: float) -> Union[float, None]:
        """
        Returns the temperature guess value according to the value
        of the property passed as argument. It is used by the root finder
        algorithm in case the return type is not `None`.

        Parameters
        ----------
        property_value : float
            value of the property

        Returns
        -------
        None
        """
        return None

    def info(self, T: float, p: float = atm,
             print_info: bool = True, n_tab: int = 0) -> Union[None, str]:
        """
        Prints the information about the property and about the correlation
        adopted to compute its value.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric
            pressure value, i.e., :math:`101325.0 Pa`
        print_info : bool, optional
            `True` to print to the console, `False` for getting the
            string. By default, `True`
        n_tab : int, optional
            Number of indentation tabs used to format the output,
            by default `0`

        Returns
        -------
        None | str
            Returns `None` if `print_info=True`, otherwise returns the content
            of the information.
        """
        property_val = self.correlation(T, p)
        if property_val < 1e-2:
            value = f"Value: {property_val:.2e} {self.units}"
        else:
            value = f"Value: {property_val:.2f} {self.units}"
        validity = ("Validity range: "
                    f"[{self.range[0]:.2f}, {self.range[1]:.2f}] K")
        corr_name = f"Correlation name: '{self.correlation_name}'"
        long_name = f"Long name: {self.long_name}"
        units = f"Units: {self.units}"
        description = ("Description:\n{:s}{:s}"
                       .format((n_tab+2)*"\t", self.description))

        all_info = "{:s}{:s}:\n".format(n_tab*"\t", self.name)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", value)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", validity)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", corr_name)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", long_name)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", units)
        all_info += "{:s}{:s}".format((n_tab+1)*"\t", description)

        if print_info:
            return print(all_info)
        return all_info

    @abstractmethod
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the property by applying the
        corresponding correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure,
            i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        exception
        """
        raise NotImplementedError(f"{type(self).__name__}.correlation "
                                  "NOT IMPLEMENTED")

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return type(self).__name__

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "lbh15"

    @property
    def is_injective(self) -> bool:
        """
        bool : `True` if the correlation is injective,
        `False` otherwise
        """
        return True

    @property
    def min(self) -> float:
        """
        float : Minimum value of the property correlation function within
        the validity range
        """
        return self.__min

    @property
    def max(self) -> float:
        """
        float : Maximum value of the property correlation function within
        the validity range
        """
        return self.__max

    @property
    def T_at_min(self) -> float:
        """
        float : Temperature value corresponding to the minimum value of the
        property correlation function within the validity range
        """
        return self.__T_at_min

    @property
    def T_at_max(self) -> float:
        """
        float : Temperature value corresponding to the maximum value of the
        property correlation function within the validity range
        """
        return self.__T_at_max

    @property
    @abstractmethod
    def range(self) -> List[float]:
        """
        List[float] : Validity range of the property correlation function
        identified by the minimum and the maximum values of the temperature
        """
        raise NotImplementedError(f"{type(self).__name__}.range "
                                  "NOT IMPLEMENTED")

    @property
    @abstractmethod
    def units(self) -> str:
        """
        str : Property unit
        """
        raise NotImplementedError(f"{type(self).__name__}.units "
                                  "NOT IMPLEMENTED")

    @property
    @abstractmethod
    def long_name(self) -> str:
        """
        str : Property long name
        """
        raise NotImplementedError(f"{type(self).__name__}.long_name "
                                  "NOT IMPLEMENTED")

    @property
    @abstractmethod
    def description(self) -> str:
        """
        str : Property description
        """
        raise NotImplementedError(f"{type(self).__name__}.description "
                                  "NOT IMPLEMENTED")



"""Module with the definition of common objects for lbh15 package"""
# KEYWORDS

# LEAD CONSTANTS
LEAD_MELTING_TEMPERATURE: float = 600.6  # [K]
LEAD_MELTING_LATENT_HEAT: float = 23.07e3  # [J/kg]
LEAD_BOILING_TEMPERATURE: float = 2021  # [K]
LEAD_VAPORISATION_HEAT: float = 858.6e3  # [J/kg]
LEAD_MOLAR_MASS: float = 0.20720  # [kg/mol]

# BISMUTH CONSTANTS
BISMUTH_MELTING_TEMPERATURE: float = 544.6  # [K]
BISMUTH_MELTING_LATENT_HEAT: float = 53.3e3  # [J/kg]
BISMUTH_BOILING_TEMPERATURE: float = 1831  # [K]
BISMUTH_VAPORISATION_HEAT: float = 856.2e3  # [J/kg]
BISMUTH_MOLAR_MASS: float = 0.20898  # [kg/mol]

# LEAD-BISMUTH-EUTECTIC CONSTANTS
LBE_MELTING_TEMPERATURE: float = 398.0  # [K]
LBE_MELTING_LATENT_HEAT: float = 38.6e3  # [J/kg]
LBE_BOILING_TEMPERATURE: float = 1927  # [K]
LBE_VAPORISATION_HEAT: float = 856.6e3  # [J/kg]
LBE_MOLAR_MASS: float = 0.55*BISMUTH_MOLAR_MASS \
    + 0.45*LEAD_MOLAR_MASS  # [kg/mol]

# OXYGEN CONSTANT
OXYGEN_MOLAR_MASS: float = 0.016  # [kg/mol]



"""Module with the definition of the *thermo-chemical*
property objects for *lead-bismuth eutectic* (*lbe*)."""
from typing import List
from typing import Union
import numpy as np
from scipy.constants import atm
from scipy.constants import R
from ..interface import PropertyInterface
from ..tch_common_interface import OxygenPartialPressureInterface
from ..tch_common_interface import MolarEnthalpyInterface
from ..tch_common_interface import MolarEntropyInterface
from ..tch_common_interface import GibbsFreeEnergyInterface
from ..lbe_properties import h
from ..._commons import LBE_BOILING_TEMPERATURE as T_b0
from ..._commons import LBE_MELTING_TEMPERATURE as T_m0
from ..._commons import LBE_MOLAR_MASS as M
from ..._commons import OXYGEN_MOLAR_MASS as M_O
from ..._decorators import range_warning


class OxygenPartialPressure(OxygenPartialPressureInterface):
    """
    *Oxygen partial pressure in liquid lbe divided by the
    Oxygen concentration in liquid lbe squared* property class.
    """
    def initialization_helper(self,
                              property_value: float) -> Union[None, float]:
        """
        Returns the temperature guess value according to the value
        of the Oxygen partial pressure in liquid lbe divided by the
        Oxygen concentration in liquid lbe squared passed as argument.
        It is used by the root finder algorithm.

        Parameters
        ----------
        property_value : float
            oxygen partial pressure in liquid lbe divided by the
            oxygen concentration in liquid lbe squared
            in :math:`[Pa /wt.\\%^2]`

        Returns
        -------
        float
            Temperature guess value in :math:`[K]`
        """
        return 1600

    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen partial pressure in liquid lbe
        divided by the Oxygen concentration in liquid lbe squared* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Oxygen partial pressure divided by Oxygen concentration
            squared in :math:`[Pa / wt.\\%^2]`
        """
        return np.power(10, 2 / 2.3 / R * (-127398 / T + 27.938))\
            * M * M / M_O / M_O * 101325

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen partial
        pressure divided by the Oxygen concentration squared
        correlation function
        """
        return [812, 1008]

    @property
    def description(self) -> str:
        """
        str : Oxygen partial pressure in liquid lbe divided by the
        Oxygen concentration in liquid lbe squared description
        """
        return f"{self.long_name} in liquid lbe"


class LeadChemicalActivity(PropertyInterface):
    """
    *Lead chemical activity* in liquid lbe property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *lead chemical activity* in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Lead chemical activity in :math:`[-]`
        """
        return 0.42206 - 63.2 / T

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "pb_a"

    @property
    def units(self) -> str:
        """
        str : Lead chemical activity unit
        """
        return "[-]"

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the lead chemical
        activity correlation function
        """
        return [399, 1173]

    @property
    def long_name(self) -> str:
        """
        str : Lead chemical activity long name
        """
        return "lead chemical activity"

    @property
    def description(self) -> str:
        """
        str : Lead chemical activity description
        """
        return f"{self.long_name} in liquid lbe"


class BismuthChemicalActivity(PropertyInterface):
    """
    *Bismuth chemical activity* in liquid lbe property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *bismuth chemical activity* in liquid lbe
        by applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Bismuth chemical activity in :math:`[-]`
        """
        return 0.53381 - 56.2 / T

    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "bi_a"

    @property
    def units(self) -> str:
        """
        str : Bismuth chemical activity unit
        """
        return "[-]"

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the bismuth chemical
        activity correlation function
        """
        return [399, 1173]

    @property
    def long_name(self) -> str:
        """
        str : Bismuth chemical activity long name
        """
        return "Bismuth chemical activity"

    @property
    def description(self) -> str:
        """
        str : Bismuth chemical activity description
        """
        return f"{self.long_name} in liquid lbe"


class MolarEnthalpy(MolarEnthalpyInterface):
    """
    Liquid lbe *molar enthalpy variation* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *molar enthalpy variation* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            molar enthalpy in :math:`[J/mol]`
        """
        return h().correlation(T, p) * M / 1000

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the molar enthalpy
        variation correlation function
        """
        return [400.0, T_b0]

    @property
    def description(self) -> str:
        """
        str : Molar enthalpy variation description
        """
        return f"{self.long_name} in liquid lbe"


class MolarEntropy(MolarEntropyInterface):
    """
    Liquid lbe *molar entropy variation* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *molar entropy variation* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            molar entropy in :math:`[J/(mol \\cdot K)]`
        """
        return M / 1000 * (164.8 * np.log(T / T_m0)
                           - 3.94e-2 * (T - T_m0)
                           + 6.25e-6 * (T * T - T_m0 * T_m0)
                           + 2.28e5 * (1 / T / T - 1 / T_m0 / T_m0))

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the molar entropy
        variation correlation function
        """
        return [400, T_b0]

    @property
    def description(self) -> str:
        """
        str : Molar entropy variation description
        """
        return f"{self.long_name} in liquid lbe"


class GibbsFreeEnergy(GibbsFreeEnergyInterface):
    """
    Liquid lbe *Gibbs free energy variation* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Gibbs free energy variation* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            Gibbs free energy in :math:`[J/mol]`
        """
        return MolarEnthalpy().correlation(T, p)\
            - T * MolarEntropy().correlation(T, p)

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Gibbs free energy
        variation correlation function
        """
        return [400, T_b0]

    @property
    def description(self) -> str:
        """
        str : Gibbs free energy description
        """
        return f"{self.long_name} in liquid lbe"



"""Module with the definition of the *diffusivity*
property objects for *lead-bismuth eutectic* (*lbe*)."""
from typing import List
import numpy as np
from scipy.constants import atm, R
from ..tch_common_interface import OxygenDiffusivityInterface
from ..tch_common_interface import IronDiffusivityInterface
from ..._decorators import range_warning


class LBEOxygenDiffusivityInterface(OxygenDiffusivityInterface):
    """
    Liquid lbe *Oxygen diffusivity* property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Oxygen diffusivity description
        """
        return f"{self.long_name} in liquid lbe"


class OxygenDiffusivityGromov1996(LBEOxygenDiffusivityInterface):
    """
    Liquid lbe *Oxygen diffusivity* property class
    implementing the correlation by *gromov1996*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-43073 / R / T) * 2.39e-6

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gromov1996"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [473, 1273]


class OxygenDiffusivityGanesan2006b(LBEOxygenDiffusivityInterface):
    """
    Liquid lbe *Oxygen diffusivity* property class
    implementing the correlation by *ganesan2006b*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Oxygen diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.exp(-69069 / R / T) * 0.154e-4

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "ganesan2006b"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen diffusivity
        correlation function
        """
        return [813, 973]


class IronDiffusivity(IronDiffusivityInterface):
    """
    Liquid lbe *Iron diffusivity* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron diffusivity* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            diffusivity in :math:`[m^2 / s]`
        """
        return np.power(10, - 2.31 - 2295 / T) * 1.0e-4

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron diffusivity
        correlation function
        """
        return [973.0, 1273.0]

    @property
    def description(self) -> str:
        """
        str : Iron diffusivity description
        """
        return f"{self.long_name} in liquid lbe"



"""Module with the definition of bismuth liquid metal object class,
i.e., Bismuth"""
import copy
from typing import Dict
from typing import List
from scipy.constants import atm
from ._commons import BISMUTH_MELTING_TEMPERATURE
from ._commons import BISMUTH_MELTING_LATENT_HEAT
from ._commons import BISMUTH_BOILING_TEMPERATURE
from ._commons import BISMUTH_VAPORISATION_HEAT
from ._commons import BISMUTH_MOLAR_MASS
from ._lbh15 import LiquidMetalInterface
from .properties.interface import PropertyInterface


class Bismuth(LiquidMetalInterface):
    """
    Class to model liquid bismuth properties either at a given temperature or
    at a given value of a specific property to choose among a list of
    available properties.

    Parameters
    ----------
    p : float, optional
        Pressure in [Pa], by default the atmospheric pressure value, i.e.,
        101325.0 Pa
    \\**kwargs : dict
        One-item dictionary that specifies the quantity which the object shall
        be initialized from. The available ones by default are:

        - **T** (float) : temperature :math:`[K]`
        - **p_s** (float) : saturation vapour pressure :math:`[Pa]`
        - **sigma** (float) : surface tension :math:`[N/m]`
        - **rho** (float) : density :math:`[kg/m^3]`
        - **alpha** (float) : thermal expansion coefficient :math:`[1/K]`
        - **u_s** (float) : speed of sound :math:`[m/s]`
        - **beta_s** (float) : isentropic compressibility :math:`[1/Pa]`
        - **cp** (float) : specific heat capacity :math:`[J/(kg \\cdot K)]`
        - **h** (float) : specific enthalpy \
            (with respect to melting point) :math:`[J/kg]`
        - **mu** (float) : dynamic viscosity :math:`[Pa \\cdot s]`
        - **r** (float) : electrical resistivity :math:`[Ohm \\cdot m]`
        - **k** (float) : thermal conductivity :math:`[W/(m \\cdot K)]`
        - **H** (float) : molar enthalpy :math:`[J/mol]`
        - **S** (float) : molar entropy :math:`[J/(mol \\cdot K)]`
        - **G** (float) : Gibbs free energy :math:`[J/mol]`
        - **fe_sol** (float) : Iron solubility :math:`[wt.\\%]`
        - **ni_sol** (float) : Nickel solubility :math:`[wt.\\%]`
        - **cr_sol** (float) : Chromium solubility :math:`[wt.\\%]`
        - **o_sol** (float) : Oxygen solubility :math:`[wt.\\%]`
        - **o_dif** (float) : Oxygen diffusivity :math:`[m^2 / s]`
        - **o_pp** (float) : Oxygen partial pressure divided by Oxygen \
            concentration squared :math:`[Pa / wt.\\%^2]`

    Example
    -------
    >>> liquid_bismuth = Bismuth(T=670.0)
    >>> liquid_bismuth.k  # [W/(m*K)]
    13.705
    """
    _default_corr_to_use: Dict[str, str] = \
        {'fe_sol': "gosse2014", 'ni_sol': "gosse2014",
         'cr_sol': "gosse2014", 'o_dif': "fitzner1980",
         'o_pp': "isecke1979"}
    _correlations_to_use: Dict[str, str] = copy.deepcopy(_default_corr_to_use)
    _roots_to_use: Dict[str, int] = {'cp': 0}
    _custom_properties_path: Dict[str, List[str]] = {}
    _available_properties_dict: Dict[str, PropertyInterface] = {}
    _available_correlations_dict: Dict[str, List[str]] = {}
    _properties_modules_list: List[str] = \
        ['lbh15.properties.bismuth_thermochemical_properties\
.solubility_in_bismuth',
         'lbh15.properties.bismuth_thermochemical_properties\
.diffusivity_in_bismuth',
         'lbh15.properties.bismuth_thermochemical_properties\
.bismuth_thermochemical',
         'lbh15.properties.bismuth_properties']

    def __init__(self, p: float = atm, **kwargs):
        self._guess = BISMUTH_BOILING_TEMPERATURE / 2.0
        super().__init__(p=p, **kwargs)

    def _set_constants(self) -> None:
        """
        Sets the class constants
        """
        self._T_m0 = BISMUTH_MELTING_TEMPERATURE
        self._Q_m0 = BISMUTH_MELTING_LATENT_HEAT
        self._T_b0 = BISMUTH_BOILING_TEMPERATURE
        self._Q_b0 = BISMUTH_VAPORISATION_HEAT
        self._M = BISMUTH_MOLAR_MASS



"""__init__ module of lbh15 package"""

__version__ = "2.1.0"
__author__ = "Daniele Panico, Daniele Tomatis, Gabriele Ottino"
__company__ = "newcleo"
__date__ = "04 April 2024"

from .lead import Lead
from .bismuth import Bismuth
from .lbe import LBE
from .properties import lead_properties
from .properties import bismuth_properties
from .properties import lbe_properties
from .properties.bismuth_thermochemical_properties import solubility_in_bismuth
from .properties.bismuth_thermochemical_properties import diffusivity_in_bismuth
from .properties.bismuth_thermochemical_properties import bismuth_thermochemical
from .properties.lbe_thermochemical_properties import solubility_in_lbe
from .properties.lbe_thermochemical_properties import diffusivity_in_lbe
from .properties.lbe_thermochemical_properties import lbe_thermochemical
from .properties.lbe_thermochemical_properties import lbe_oxygen_limits
from .properties.lead_thermochemical_properties import solubility_in_lead
from .properties.lead_thermochemical_properties import diffusivity_in_lead
from .properties.lead_thermochemical_properties import lead_thermochemical
from .properties.lead_thermochemical_properties import lead_oxygen_limits


"""Module with the definition of common property interfaces
for thermochemical properties"""
from .interface import PropertyInterface


class OxygenDiffusivityInterface(PropertyInterface):
    """
    Liquid metal *Oxygen diffusivity* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "o_dif"

    @property
    def units(self) -> str:
        """
        str : Oxygen diffusivity unit
        """
        return "[m^2.s^-1]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen diffusivity long name
        """
        return "oxygen diffusivity"


class IronDiffusivityInterface(PropertyInterface):
    """
    Liquid metal *Iron diffusivity* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "fe_dif"

    @property
    def units(self) -> str:
        """
        str : Iron diffusivity unit
        """
        return "[m^2.s^-1]"

    @property
    def long_name(self) -> str:
        """
        str : Iron diffusivity long name
        """
        return "iron diffusivity"


class OxygenPartialPressureInterface(PropertyInterface):
    """
    *Oxygen partial pressure in liquid metal divided by the
    Oxygen concentration in liquid metal squared* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "o_pp"

    @property
    def units(self) -> str:
        """
        str : Oxygen partial pressure in liquid metal divided by the
        Oxygen concentration in liquid metal squared unit
        """
        return "[Pa.wt.%^-2]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen partial pressure in liquid metal divided by the
        Oxygen concentration in liquid metal squared long name
        """
        return ("Oxygen partial pressure divided by the"
                " oxygen concentration squared")


class MolarEnthalpyInterface(PropertyInterface):
    """
    Liquid metal *molar enthalpy variation* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "H"

    @property
    def units(self) -> str:
        """
        str : Molar enthalpy variation unit
        """
        return "[J.mol^-1]"

    @property
    def long_name(self) -> str:
        """
        str : Molar enthalpy variation long name
        """
        return "molar enthalpy variation"


class MolarEntropyInterface(PropertyInterface):
    """
    Liquid metal *molar entropy variation* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "S"

    @property
    def units(self) -> str:
        """
        str : Molar entropy variation unit
        """
        return "[J/(mol.K)]"

    @property
    def long_name(self) -> str:
        """
        str : Molar entropy variation long name
        """
        return "molar entropy variation"


class GibbsFreeEnergyInterface(PropertyInterface):
    """
    Liquid metal *Gibbs free energy variation* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "G"

    @property
    def units(self) -> str:
        """
        str : Gibbs free energy unit
        """
        return "[J/mol]"

    @property
    def long_name(self) -> str:
        """
        str : Gibbs free energy long name
        """
        return "Gibbs free energy variation"


class IronSolubilityInterface(PropertyInterface):
    """
    Liquid metal *Iron solubility* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "fe_sol"

    @property
    def units(self) -> str:
        """
        str : Iron solubility unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Iron solubility long name
        """
        return "iron solubility"


class NickelSolubilityInterface(PropertyInterface):
    """
    Liquid metal *Nickel solubility* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "ni_sol"

    @property
    def units(self) -> str:
        """
        str : Nickel solubility unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Nickel solubility long name
        """
        return "nickel solubility"


class ChromiumSolubilityInterface(PropertyInterface):
    """
    Liquid metal *Chromium solubility* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "cr_sol"

    @property
    def units(self) -> str:
        """
        str : Chromium solubility unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Chromium solubility long name
        """
        return "chromium solubility"


class OxygenSolubilityInterface(PropertyInterface):
    """
    Liquid metal *Oxygen solubility* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "o_sol"

    @property
    def units(self) -> str:
        """
        str : Oxygen solubility unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen solubility long name
        """
        return "oxygen solubility"


class LowerLimitSaturationIronInterface(PropertyInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid metal considering
    *Iron at its saturation concentration* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_fe_sat"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit for"
                "iron at its saturation concentration")


class LowerLimitSaturationChromiumInterface(PropertyInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid metal considering
    *Chromium at its saturation concentration* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_cr_sat"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit for"
                " chromium at its saturation concentration")


class LowerLimitSaturationNickelInterface(PropertyInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid metal considering
    *Nickel at its saturation concentration* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_ni_sat"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit for"
                " nickel at its saturation concentration")


class LowerLimitSaturationSiliconInterface(PropertyInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid metal considering
    *Silicon at its saturation concentration* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_si_sat"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit for"
                " silicon at its saturation concentration")


class LowerLimitSaturationAluminiumInterface(PropertyInterface):
    """
    *Lower limit of Oxygen concentration* to promote a
    protective oxide film in liquid metal considering
    *Aluminium at its saturation concentration* property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_al_sat"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit for"
                " aluminium at its saturation concentration")


class LowerLimitChromiumInterface(PropertyInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Chromium concentration*
    raised to :math:`2/3` in liquid metal property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_cr"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit times"
                " chromium concentration raised to 2/3")


class LowerLimitNickelInterface(PropertyInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Nickel concentration*
    in liquid metal property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_ni"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit times"
                " nickel concentration")


class LowerLimitIronInterface(PropertyInterface):
    """
    *Lower limit of oxygen concentration* to promote a
    protective oxide film times the *Iron concentration*
    raised to :math:`3/4` in liquid metal property abstract class.
    """
    @property
    def name(self) -> str:
        """
        str : Name of the property
        """
        return "lim_fe"

    @property
    def units(self) -> str:
        """
        str : Oxygen concentration lower limit unit
        """
        return "[wt.%]"

    @property
    def long_name(self) -> str:
        """
        str : Oxygen concentration lower limit long name
        """
        return ("Oxygen concentration lower limit times"
                " iron concentration raised to 3/4")



# This test is used to check initialization from properties
import unittest
import sys
import os
import inspect
sys.path.insert(0, os.path.abspath('..'))
from lbh15.properties.interface import PropertyInterface
from lbh15 import Lead
from lbh15 import lead_properties
from scipy.constants import convert_temperature


def load_prop(module_name):
    propertyObjectList = []

    def is_valid(obj):
        return inspect.isclass(obj) and obj is not PropertyInterface \
            and not inspect.isabstract(obj) \
            and issubclass(obj, PropertyInterface)

    for _, obj in inspect.getmembers(sys.modules[module_name], is_valid):
        propertyObjectList.append(obj())
    return propertyObjectList


tol = 8
Ts = [330.0] + list((range(350, 1750, 50))) + [1745.0]
Ts = list(map(lambda x: convert_temperature(x, 'C', 'K'), Ts))
leadPs = []
for T in Ts:
    leadPs.append(Lead(T=T))

# Compute temperature discriminants for cp root index identification
cp_sobolev2011 = lead_properties.cp_sobolev2011()
cp_sobolev2011.compute_bounds()
T_change_sobolev2011 = cp_sobolev2011.T_at_min
cp_gurvich1991 = lead_properties.cp_gurvich1991()
cp_gurvich1991.compute_bounds()
T_change_gurvich1991 = cp_gurvich1991.T_at_min


class LeadTester(unittest.TestCase):

    def test_init_fromX(self):
        for leadP in leadPs:
            properties = load_prop('lbh15.properties.lead_properties')
            for prop in properties:
                name = prop.name
                if name == 'cp' and\
                        ((prop.correlation_name == 'sobolev2011' \
                        and leadP.T > T_change_sobolev2011) \
                        or \
                        (prop.correlation_name == 'gurvich1991' \
                        and leadP.T > T_change_gurvich1991)):
                    Lead.set_root_to_use('cp', 1)
                val = getattr(leadP, name)
                init_dict = {name: val}
                fromX = Lead(**init_dict)
                self.assertAlmostEqual(leadP.T, fromX.T, tol, name+" FAILED")


class LeadSolubilityTester(unittest.TestCase):

    def test_init_fromX(self):
        for leadP in leadPs:
            properties = load_prop('lbh15.properties.lead_thermochemical_properties.solubility_in_lead')
            for prop in properties:
                name = prop.name
                val = getattr(leadP, name)
                init_dict = {name: val}
                fromX = Lead(**init_dict)
                self.assertAlmostEqual(leadP.T, fromX.T, tol, name+" FAILED")


class LeadDiffusivityTester(unittest.TestCase):

    def test_init_fromX(self):
        for leadP in leadPs:
            properties = load_prop('lbh15.properties.lead_thermochemical_properties.diffusivity_in_lead')
            for prop in properties:
                name = prop.name
                val = getattr(leadP, name)
                init_dict = {name: val}
                fromX = Lead(**init_dict)
                self.assertAlmostEqual(leadP.T, fromX.T, tol, name+" FAILED")


class LeadThermochemicalTester(unittest.TestCase):

    def test_init_fromX(self):
        for leadP in leadPs:
            properties = load_prop('lbh15.properties.lead_thermochemical_properties.lead_thermochemical')
            for prop in properties:
                name = prop.name
                val = getattr(leadP, name)
                init_dict = {name: val}
                fromX = Lead(**init_dict)
                self.assertAlmostEqual(leadP.T, fromX.T, tol, name+" FAILED")


class LeadLimitsTester(unittest.TestCase):

    def test_init_fromX(self):
        for leadP in leadPs:
            properties = load_prop('lbh15.properties.lead_thermochemical_properties.lead_oxygen_limits')
            for prop in properties:
                name = prop.name
                val = getattr(leadP, name)
                init_dict = {name: val}
                fromX = Lead(**init_dict)
                self.assertAlmostEqual(leadP.T, fromX.T, tol, name+" FAILED")


if __name__ == "__main__":
    unittest.main()



# This test is used to check initialization from properties
import unittest
import sys
import os
import inspect
sys.path.insert(0, os.path.abspath('..'))
from lbh15.properties.interface import PropertyInterface
from lbh15 import LBE
from scipy.constants import convert_temperature


def load_prop(module_name):
    propertyObjectList = []

    def is_valid(obj):
        return inspect.isclass(obj) and obj is not PropertyInterface \
            and not inspect.isabstract(obj) \
            and issubclass(obj, PropertyInterface)

    for _, obj in inspect.getmembers(sys.modules[module_name], is_valid):
        propertyObjectList.append(obj())
    return propertyObjectList


tol = 8
Ts = [125] + list(range(150, 1700, 50))
Ts = list(map(lambda x: convert_temperature(x, 'C', 'K'), Ts))
lbePs = []
for T in Ts:
    lbePs.append(LBE(T=T))


class LBETester(unittest.TestCase):

    def test_init_fromX(self):
        for lbeP in lbePs:
            properties = load_prop('lbh15.properties.lbe_properties')
            for prop in properties:
                name = prop.name
                if name in LBE.roots_to_use().keys():
                    continue
                val = getattr(lbeP, name)
                init_dict = {name: val}
                fromX = LBE(**init_dict)
                self.assertAlmostEqual(lbeP.T, fromX.T, tol, name+" FAILED")


class LBESolubilityTester(unittest.TestCase):

    def test_init_fromX(self):
        for lbeP in lbePs:
            properties = load_prop('lbh15.properties.lbe_thermochemical_properties.solubility_in_lbe')
            for prop in properties:
                name = prop.name
                val = getattr(lbeP, name)
                init_dict = {name: val}
                fromX = LBE(**init_dict)
                self.assertAlmostEqual(lbeP.T, fromX.T, tol, name+" FAILED")


class LBEDiffusivityTester(unittest.TestCase):

    def test_init_fromX(self):
        for lbeP in lbePs:
            properties = load_prop('lbh15.properties.lbe_thermochemical_properties.diffusivity_in_lbe')
            for prop in properties:
                name = prop.name
                val = getattr(lbeP, name)
                init_dict = {name: val}
                fromX = LBE(**init_dict)
                self.assertAlmostEqual(lbeP.T, fromX.T, tol, name+" FAILED")


class LBEThermochemicalTester(unittest.TestCase):

    def test_init_fromX(self):
        for lbeP in lbePs:
            properties = load_prop('lbh15.properties.lbe_thermochemical_properties.lbe_thermochemical')
            for prop in properties:
                name = prop.name
                val = getattr(lbeP, name)
                init_dict = {name: val}
                fromX = LBE(**init_dict)
                self.assertAlmostEqual(lbeP.T, fromX.T, tol, name+" FAILED")


class LBELimitsTester(unittest.TestCase):

    def test_init_fromX(self):
        for lbeP in lbePs:
            properties = load_prop('lbh15.properties.lbe_thermochemical_properties.lbe_oxygen_limits')
            for prop in properties:
                name = prop.name
                val = getattr(lbeP, name)
                init_dict = {name: val}
                fromX = LBE(**init_dict)
                self.assertAlmostEqual(lbeP.T, fromX.T, tol, name+" FAILED")


if __name__ == "__main__":
    unittest.main()



# This test is used to check lbh15 static methods
import unittest
import sys
import os
import json
import inspect
sys.path.insert(0, os.path.abspath('..'))
import lbh15
from lbh15.properties.interface import PropertyInterface


def load_prop(module_name):
    propertyObjectList = []
    module = module_name
    for name, obj in inspect.getmembers(sys.modules[module]):
        if (inspect.isclass(obj) and obj is not PropertyInterface
                and not inspect.isabstract(obj)):
            if issubclass(obj, PropertyInterface):
                propertyObjectList.append(obj())
    return propertyObjectList


def get_val(prop_object, what):
    key = (prop_object.name + "_"
           + prop_object.correlation_name + "_"
           + prop_object.description)
    key = key.replace(" ", "_")
    return file_bounds[key][what], key


with open("properties_bounds.json", "r") as json_file:
    file_bounds = json.load(json_file)


tol = 6


class LeadTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lead_properties')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LeadSolubilityTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lead_thermochemical_properties.solubility_in_lead')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LeadDiffusivityTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lead_thermochemical_properties.diffusivity_in_lead')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LeadThermochemicalTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lead_thermochemical_properties.lead_thermochemical')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LeadLimitsTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lead_thermochemical_properties.lead_oxygen_limits')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class BismuthTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.bismuth_properties')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class BismuthSolubilityTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.bismuth_thermochemical_properties.solubility_in_bismuth')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class BismuthDiffusivityTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.bismuth_thermochemical_properties.diffusivity_in_bismuth')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class BismuthThermochemicalTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.bismuth_thermochemical_properties.bismuth_thermochemical')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LBETester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lbe_properties')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LBESolubilityTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lbe_thermochemical_properties.solubility_in_lbe')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LBEDiffusivityTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lbe_thermochemical_properties.diffusivity_in_lbe')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LBEThermochemicalTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lbe_thermochemical_properties.lbe_thermochemical')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


class LBELimitsTester(unittest.TestCase):

    properties = load_prop('lbh15.properties.lbe_thermochemical_properties.lbe_oxygen_limits')

    def test_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "min")
            self.assertAlmostEqual(val, prop.min, tol, key+" FAILED")

    def test_T_at_min(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_min")
            self.assertAlmostEqual(val, prop.T_at_min, tol, key+" FAILED")

    def test_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "max")
            self.assertAlmostEqual(val, prop.max, tol, key+" FAILED")

    def test_T_at_max(self):
        for prop in self.properties:
            prop.compute_bounds()
            val, key = get_val(prop, "T_at_max")
            self.assertAlmostEqual(val, prop.T_at_max, tol, key+" FAILED")


if __name__ == "__main__":
    unittest.main()



"""Module with the definition of the *solubility*
property objects for *bismuth*."""
from typing import List
from typing import Union
import numpy as np
from scipy.constants import atm
from ..tch_common_interface import IronSolubilityInterface
from ..tch_common_interface import NickelSolubilityInterface
from ..tch_common_interface import ChromiumSolubilityInterface
from ..tch_common_interface import OxygenSolubilityInterface
from ..._decorators import range_warning


class BismuthIronSolubilityInterface(IronSolubilityInterface):
    """
    Liquid bismuth *Iron solubility* property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Iron solubility description
        """
        return f"{self.long_name} in liquid bismuth"


class IronSolubilityGosse2014(BismuthIronSolubilityInterface):
    """
    Liquid bismuth *Iron solubility* property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.20-3930/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron solubility
        correlation function
        """
        return [545.0, 1173.0]


class IronSolubilityMassalski1990(BismuthIronSolubilityInterface):
    """
    Liquid bismuth *Iron solubility* property class
    implementing the correlation by *massalski1990*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.18-3980/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "massalski1990"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron solubility
        correlation function
        """
        return [973.0, 1173.0]


class IronSolubilityWeeks1998(BismuthIronSolubilityInterface):
    """
    Liquid bismuth *Iron solubility* property class
    implementing the correlation by *weeks1998*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Iron solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 1.832-3589/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "weeks1998"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Iron solubility
        correlation function
        """
        return [713.0, 998.0]


class BismuthNickelSolubilityInterface(NickelSolubilityInterface):
    """
    Liquid bismuth *Nickel solubility* property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Nickel solubility description
        """
        return f"{self.long_name} in liquid bismuth"


class NickelSolubilityWeeks1998(BismuthNickelSolubilityInterface):
    """
    Liquid bismuth *Nickel solubility* property class
    implementing the correlation by *weeks1998*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Nickel solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.61-1538/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "weeks1998"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Nickel solubility
        correlation function
        """
        return [723.0, 903.0]


class NickelSolubilityGosse2014(BismuthNickelSolubilityInterface):
    """
    Liquid bismuth *Nickel solubility* property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> Union[float, np.ndarray]:
        """
        Returns the value of the *Nickel solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.where(T <= 738, np.power(10, 3.81-2429/T),
                        np.where(T <= 918, np.power(10, 2.05-1131/T),
                                 np.power(10, 1.35-484/T)))[()]

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Nickel solubility
        correlation function
        """
        return [543.0, 1173.0]


class BismuthChromiumSolubilityInterface(ChromiumSolubilityInterface):
    """
    Liquid bismuth *Chromium solubility* property abstract class.
    """
    @property
    def description(self) -> str:
        """
        str : Chromium solubility description
        """
        return f"{self.long_name} in liquid bismuth"


class ChromiumSolubilityVenkatraman1988(BismuthChromiumSolubilityInterface):
    """
    Liquid bismuth *Chromium solubility* property class
    implementing the correlation by *venkatraman1988*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.34-3610/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "venkatraman1988"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [658.0, 901.0]


class ChromiumSolubilityWeeks1998(BismuthChromiumSolubilityInterface):
    """
    Liquid bismuth *Chromium solubility* property class
    implementing the correlation by *weeks1998*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.5-3717/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "weeks1998"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [663.0, 998.0]


class ChromiumSolubilityGosse2014(BismuthChromiumSolubilityInterface):
    """
    Liquid bismuth *Chromium solubility* property class
    implementing the correlation by *gosse2014*.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> float:
        """
        Returns the value of the *Chromium solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.power(10, 2.34-3610/T)

    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "gosse2014"

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Chromium solubility
        correlation function
        """
        return [545.0, 1773.0]


class OxygenSolubility(OxygenSolubilityInterface):
    """
    Liquid bismuth *Oxygen solubility* property class.
    """
    @range_warning
    def correlation(self, T: float, p: float = atm,
                    verbose: bool = False) -> Union[float, np.ndarray]:
        """
        Returns the value of the *Oxygen solubility* by
        applying the property correlation.

        Parameters
        ----------
        T : float
            Temperature in :math:`[K]`
        p : float, optional
            Pressure in :math:`[Pa]`, by default the atmospheric pressure
            value, i.e., :math:`101325.0 Pa`
        verbose : bool, optional
            `True` to tell the decorator to print a warning message in case of
            range check failing, `False` otherwise. By default, `False`

        Returns
        -------
        float:
            solubility in :math:`[wt.\\%]`
        """
        return np.where(T <= 1002, np.power(10, 2.30-4066/T),
                        np.power(10, 3.04-4810/T))[()]

    @property
    def range(self) -> List[float]:
        """
        List[float] : Temperature validity range of the Oxygen solubility
        correlation function
        """
        return [573.0, 1573.0]

    @property
    def description(self) -> str:
        """
        str : Oxygen solubility description
        """
        return f"{self.long_name} in liquid bismuth"



#!/usr/bin/python3

from setuptools import setup
from setuptools import find_packages
import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_info(rel_path, info):
    for line in read(rel_path).splitlines():
        if line.startswith(f'__{info}__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError(f"Unable to find {info} string.")


if __name__ == '__main__':

    setup(
        name='lbh15',
        version=get_info('lbh15/__init__.py', 'version'),
        packages=find_packages(),
        include_package_data=True,
        author=get_info('lbh15/__init__.py', 'author'),
        author_email='daniele.panico@newcleo.com, daniele.tomatis@newcleo.com',
        description='Python implementation of liquid metal properties from '
                    'Handbook on Lead-bismuth Eutectic Alloy and '
                    'Lead Properties, Materials Compatibility, '
                    'Thermal-hydraulics and Technologies',
        long_description_content_type='text/x-rst',
        long_description=open('README.rst', 'r').read(),
        license='lgpl v3',
        python_requires='>=3.8.10',
        install_requires=['scipy>=1.8.1', 'numpy>=1.22.3', 'sphinx>=6.2.1',
                          'sphinx-rtd-theme>=1.3.0', 'myst-parser>=1.0.0', 'sphinxcontrib-bibtex>=2.5.0'],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    )

"""
Developers memo for release:
    1. Update package information (on master):
        - Change version in lbh15/__init__.py
        - Change date in lbh15/__init__.py
        - Change date in README.rst
        - Update the "docs/source/release_notes.rst"
        - Update the CHANGELOG.rst file with the features introduced in the last release
        - Update author list in README.rst

    2. Test the package build and install on test-pypi:
        a. Change the name of the package in setup.py and pyproject.toml in 'lbh15-test'
        b. python3 setup.py sdist                     # in project directory
        c. twine check dist/*                         # check that package is ok
        d. twine upload --repository testpypi dist/*  # verify upload is successful on test-PyPI
        e. crete a python3 virtual enviroment and verify the correct package installation form test-pypi:
            python3 -m venv venv_install
            source venv_install/bn/activate
            pip3 install wheel
            pip3 install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple lbh15-test
            cd tests/
            <execute tests>
        f. If every thing is ok change back the package name in 'lbh15'

    3. Create tag (on master):
        a. commit previous release info modifications
        b. git tag v<version>
        c. git push <origin_name> <tag_name>
        d. git diff --stat=200 <previous_tag> <tag_name> > log.diff.<tag_name>
        e. You should see the tag on github page as well. Then on github do:
            - https://github.com/newcleo-dev-team/lbh15/releases
            - Click on "Draft a new release"
            - Choose the tag
            - Insert the title name as "Release v<version>"
            - In the description write:
                Changelog between v<previous varsion> and v<version>

                ```
                < content of  log.diff.<tag_name> >
                ```

    4. Upload the package on PyPI:
        a. repeat step 2.b and 2.c
        b. twine upload dist/*
"""



import unittest
import sys
import os
import io
from typing import Union
sys.path.insert(0, os.path.abspath('..'))
from lbh15 import Lead
from lbh15 import LBE
from lbh15 import Bismuth

lead_custom_property_path = os.getcwd() \
    + '/custom_properties/lead_properties.py'
lbe_custom_property_path = os.getcwd() \
    + '/custom_properties/lbe_properties.py'
bismuth_custom_property_path = os.getcwd() \
    + '/custom_properties/bismuth_properties.py'
Lead.set_custom_properties_path(lead_custom_property_path)
LBE.set_custom_properties_path(lbe_custom_property_path)
Bismuth.set_custom_properties_path(bismuth_custom_property_path)


def infoOf(metal_object: Union[Lead, LBE, Bismuth]) -> str:
    """
    Function for capturing the printed info string of
    the liquid metal object passed as argument
    """
    # Freeze the stdout so far
    prev_stdout = sys.stdout
    # Redirect to the stdout the I/O flux
    sys.stdout = buffer = io.StringIO()
    metal_object.rho_info()
    # Restore the right stdout
    sys.stdout = prev_stdout
    # Return the string printed in the meanwhile
    return buffer.getvalue()


##################
# Expected results
# Lead
lead_ref_string = "rho:\n"
lead_ref_string += "\tValue: 9240.00 [kg/m^3]\n"
lead_ref_string += "\tValidity range: [700.00, 1900.00] K\n"
lead_ref_string += "\tCorrelation name: 'custom2022'\n"
lead_ref_string += "\tLong name: custom density\n"
lead_ref_string += "\tUnits: [kg/m^3]\n"
lead_ref_string += "\tDescription:\n"
lead_ref_string += "\t\tLiquid Lead custom density\n"
# LBE
lbe_ref_string = "rho:\n"
lbe_ref_string += "\tValue: 9160.00 [kg/m^3]\n"
lbe_ref_string += "\tValidity range: [800.00, 2000.00] K\n"
lbe_ref_string += "\tCorrelation name: 'custom2022'\n"
lbe_ref_string += "\tLong name: custom density\n"
lbe_ref_string += "\tUnits: [kg/m^3]\n"
lbe_ref_string += "\tDescription:\n"
lbe_ref_string += "\t\tLiquid LBE custom density\n"
# Bismuth
bismuth_ref_string = "rho:\n"
bismuth_ref_string += "\tValue: 9080.00 [kg/m^3]\n"
bismuth_ref_string += "\tValidity range: [900.00, 2100.00] K\n"
bismuth_ref_string += "\tCorrelation name: 'custom2022'\n"
bismuth_ref_string += "\tLong name: custom density\n"
bismuth_ref_string += "\tUnits: [kg/m^3]\n"
bismuth_ref_string += "\tDescription:\n"
bismuth_ref_string += "\t\tLiquid Bismuth custom density\n"


class CustomPropertiesTester(unittest.TestCase):

    def test_custom_rho_lead(self):
        Lead.set_correlation_to_use('rho', 'custom2022')
        liquid_lead = Lead(T=1800)
        self.assertEqual(infoOf(liquid_lead), lead_ref_string, "Lead \
                         properties not loaded properly!")

    def test_custom_rho_lbe(self):
        LBE.set_correlation_to_use('rho', 'custom2022')
        liquid_lbe = LBE(T=1800)
        self.assertEqual(infoOf(liquid_lbe), lbe_ref_string, "LBE \
                         properties not loaded properly!")

    def test_custom_rho_bismuth(self):
        Bismuth.set_correlation_to_use('rho', 'custom2022')
        liquid_bismuth = Bismuth(T=1800)
        self.assertEqual(infoOf(liquid_bismuth), bismuth_ref_string, "Bismuth \
                         properties not loaded properly!")

    def test_T_double(self):
        liquid_lead = Lead(T=1000)
        self.assertEqual(2000, liquid_lead.T_double)


if __name__ == "__main__":
    unittest.main()



"""Module with the definition of common property interfaces
for thermophysical properties"""
from .interface import PropertyInterface


class SaturationVapourPressureInterface(PropertyInterface):
    """
    Liquid metal *saturation vapour pressure* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Saturation vapour pressure unit
        """
        return "[Pa]"

    @property
    def long_name(self) -> str:
        """
        str : Saturation vapour pressure long name
        """
        return "saturation vapour pressure"


class SurfaceTensionInterface(PropertyInterface):
    """
    Liquid metal *surface tension* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Surface tension unit
        """
        return "[N/m]"

    @property
    def long_name(self) -> str:
        """
        str : Surface tension long name
        """
        return "surface tension"


class DensityInterface(PropertyInterface):
    """
    Liquid metal *density* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Density unit
        """
        return "[kg/m^3]"

    @property
    def long_name(self) -> str:
        """
        str : Density long name
        """
        return "density"


class ThermalExpansionInterface(PropertyInterface):
    """
    Liquid metal *thermal expansion coefficient* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Thermal expansion coefficient unit
        """
        return "[1/K]"

    @property
    def long_name(self) -> str:
        """
        str : Thermal expansion coefficient long name
        """
        return "thermal expansion coefficient"


class SpeedOfSoundInterface(PropertyInterface):
    """
    Liquid metal *sound velocity* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Sound velocity unit
        """
        return "[m/s]"

    @property
    def long_name(self) -> str:
        """
        str : Sound velocity long name
        """
        return "sound velocity"


class IsentropicCompressibilityInterface(PropertyInterface):
    """
    Liquid metal *isentropic compressibility* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Isentropic compressibility unit
        """
        return "[1/Pa]"

    @property
    def long_name(self) -> str:
        """
        str : Isentropic compressibility long name
        """
        return "isentropic compressibility"


class SpecificHeatInterface(PropertyInterface):
    """
    Liquid metal *specific heat capacity* property abstract class.
    """
    @property
    def correlation_name(self) -> str:
        """
        str : Name of the correlation
        """
        return "imbeni1998"

    @property
    def is_injective(self) -> bool:
        """
        bool : `True` if the correlation is injective,
        `False` otherwise.
        """
        return False

    @property
    def units(self) -> str:
        """
        str : Specific heat capacity unit
        """
        return "[J/(kg*K)]"

    @property
    def long_name(self) -> str:
        """
        str : Specific heat capacity long name
        """
        return "specific heat capacity"


class SpecificEnthalpyInterface(PropertyInterface):
    """
    Liquid metal *specific enthalpy* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Specific enthalpy unit
        """
        return "[J/kg]"

    @property
    def long_name(self) -> str:
        """
        str : Specific enthalpy long name
        """
        return "specific enthalpy"


class DynamicViscosityInterface(PropertyInterface):
    """
    Liquid metal *dynamic viscosity* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Dynamic viscosity unit
        """
        return "[Pa*s]"

    @property
    def long_name(self) -> str:
        """
        str : Dynamic viscosity long name
        """
        return "dynamic viscosity"


class ElectricalResistivityInterface(PropertyInterface):
    """
    Liquid metal *electrical resistivity* property abstract class.
    """
    @property
    def units(self) -> str:
        """
        str : Electrical resistivity unit
        """
        return "[Ohm*m]"

    @property
    def long_name(self) -> str:
        """
        str : Electrical resistivity long name
        """
        return "electrical resistivity"


class ThermalConductivityInterface(PropertyInterface):
    """
    Liquid metal eutectic *thermal conductivity* property class.
    """

    @property
    def units(self) -> str:
        """
        str : Thermal conductivity unit
        """
        return "[W/(m*K)]"

    @property
    def long_name(self) -> str:
        """
        str : Thermal conductivity long name
        """
        return "thermal conductivity"
